from nyckel import (
    TabularClassificationFunction,
    ClassificationLabel,
    TabularClassificationSample,
    ClassificationAnnotation,
    ClassificationPrediction,
)
import time
import pytest


def test_end_to_end(tabular_classification_function):
    func: TabularClassificationFunction = tabular_classification_function

    func.create_labels([ClassificationLabel(name="happy"), ClassificationLabel(name="sad")])

    func.create_samples(
        [
            TabularClassificationSample(
                data={"firstname": "Adam", "lastname": "Adame"}, annotation=ClassificationAnnotation(label_name="happy")
            ),
            TabularClassificationSample(
                data={"firstname": "Bo", "lastname": "Boden"}, annotation=ClassificationAnnotation(label_name="happy")
            ),
            TabularClassificationSample(
                data={"firstname": "Caren", "lastname": "Carrot"}, annotation=ClassificationAnnotation(label_name="sad")
            ),
            TabularClassificationSample(
                data={"firstname": "Dick", "lastname": "Denali"}, annotation=ClassificationAnnotation(label_name="sad")
            ),
        ]
    )

    fields = func._field_handler.list_fields()
    assert len(fields) == 2
    assert set([field.name for field in fields]) == set(["firstname", "lastname"])
    time.sleep(2)
    print(func.list_samples())
    assert len(func.list_samples()) == 4
    assert len(func.list_labels()) == 2

    while not func.has_trained_model():
        print("No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    prediction = func({"firstname": "Eric", "lastname": "Ebony"})
    assert isinstance(prediction, ClassificationPrediction)

    predictions = func.invoke(
        [{"firstname": "Eric", "lastname": "Ebony"}, {"firstname": "Frank", "lastname": "Froggy"}]
    )
    assert isinstance(predictions[0], ClassificationPrediction)


def test_field_creation(tabular_classification_function):
    func: TabularClassificationFunction = tabular_classification_function

    func.create_samples([TabularClassificationSample(data={"firstname": "Adam"})])
    fields = func._field_handler.list_fields()
    assert len(fields) == 1
    assert fields[0].type == "Text"
    func.create_samples([TabularClassificationSample(data={"age": 32})])
    fields = func._field_handler.list_fields()
    assert len(fields) == 2
    age_field = [field for field in fields if field.name == "age"][0]
    assert age_field.type == "Number"

    with pytest.raises(ValueError):
        # Can't assign a Text value to a Number field.
        func.create_samples([TabularClassificationSample(data={"age": "Twelve"})])



import time
import os
import pytest
import numpy as np
from nyckel import (
    ClassificationAnnotation,
    ClassificationLabel,
    ClassificationPrediction,
    ImageClassificationSample,
    ImageClassificationFunction,
)
from nyckel.functions.classification.image_classification import ImageDecoder
import numpy as np
from conftest import get_test_user, make_random_image

user = get_test_user()


def test_samples(image_classification_function):
    func: ImageClassificationFunction = image_classification_function
    label = ClassificationLabel(name="Nice")
    func.create_labels([label])
    time.sleep(1)

    images = [make_random_image(), make_random_image()]
    sample = ImageClassificationSample(data=images[0], annotation=ClassificationAnnotation(label_name="Nice"))
    sample_ids = func.create_samples([sample])
    sample_back = func.read_sample(sample_ids[0])
    assert ImageDecoder().to_image(sample_back.data).size == ImageDecoder().to_image(sample.data).size

    sample = ImageClassificationSample(
        data=images[1], annotation=ClassificationAnnotation(label_name="Nice"), external_id="my_external_id"
    )
    sample_ids = func.create_samples([sample])
    sample_back = func.read_sample(sample_ids[0])
    assert sample_back.external_id == sample.external_id

    # Check that list_samples work
    samples_back = func.list_samples()
    assert len(samples_back) == 2

    # Check delete
    func.delete_sample(sample_ids[0])
    time.sleep(1)
    samples_back = func.list_samples()
    assert len(samples_back) == 1


def test_end_to_end(image_classification_function):
    func: ImageClassificationFunction = image_classification_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)

    samples = [
        ImageClassificationSample(data=make_random_image(), annotation=ClassificationAnnotation(label_name="Nice")),
        ImageClassificationSample(data=make_random_image(), annotation=ClassificationAnnotation(label_name="Nice")),
        ImageClassificationSample(data=make_random_image(), annotation=ClassificationAnnotation(label_name="Boo")),
        ImageClassificationSample(data=make_random_image(), annotation=ClassificationAnnotation(label_name="Boo")),
        ImageClassificationSample(data=make_random_image()),
    ]

    func.create_samples(samples)

    while not func.has_trained_model():
        print("No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    assert isinstance(func(make_random_image()), ClassificationPrediction)

    assert len(func.invoke([make_random_image(), make_random_image(), make_random_image()])) == 3

    returned_samples = func.list_samples()
    assert len(samples) == len(returned_samples)
    assert isinstance(returned_samples[0], ImageClassificationSample)


@pytest.mark.skipif(
    user.server_url == "http://localhost:5000",
    reason="Integrity of image on copy is only implemented for images hosted on S3.",
)
def test_image_integrity_on_copy(image_classification_function):
    func: ImageClassificationFunction = image_classification_function
    local_filepath = os.path.abspath("tests/flower.jpg")

    # When Nyckel first receives an image, it resizes and recodes it.
    first_sample_id = func.create_samples([ImageClassificationSample(data=local_filepath, external_id="v0")])[0]
    first_sample = func.read_sample(first_sample_id)

    # Here we point back to a nyckel owned url, so the images bytes are stored exactly as posted.
    second_sample_id = func.create_samples([ImageClassificationSample(data=first_sample.data, external_id="v1")])[0]
    second_sample = func.read_sample(second_sample_id)

    decoder = ImageDecoder()
    assert np.array_equal(np.array(decoder.to_image(first_sample.data)), np.array(decoder.to_image(second_sample.data)))



import concurrent.futures
from typing import Dict, List, Tuple

import requests
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from nyckel.config import NBR_CONCURRENT_REQUESTS


class ParallelPoster:
    def __init__(self, session: requests.Session, endpoint: str):
        self._session = session
        self._endpoint = endpoint

    def _post_as_json(self, data: Dict) -> requests.Response:
        response = self._session.post(self._endpoint, json=data)
        return response

    def __call__(self, bodies: List[Dict]) -> List[requests.Response]:
        if len(bodies) == 0:
            return []
        responses = [requests.Response()] * len(bodies)
        n_workers = min(len(bodies), NBR_CONCURRENT_REQUESTS)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
            index_by_future = {executor.submit(self._post_as_json, body): index for index, body in enumerate(bodies)}
            for future in tqdm(concurrent.futures.as_completed(index_by_future), desc=f"Posting to {self._endpoint}"):
                index = index_by_future[future]
                body = bodies[index]
                response = future.result()
                if not response.status_code == 200:
                    raise ValueError(
                        f"Posting {body} to {self._endpoint} failed with {response.status_code=} {response.text=}"
                    )
                responses[index] = response
        return responses


class ParallelDeleter:
    def __init__(self, session: requests.Session, endpoint: str):
        self._session = session
        self._endpoint = endpoint

    def _delete_one(self, asset_id: str) -> requests.Response:
        response = self._session.delete(f"{self._endpoint}/{asset_id}")
        return response

    def __call__(self, asset_ids: List[str]) -> List[requests.Response]:
        responses = [requests.Response()] * len(asset_ids)
        n_workers = min(len(asset_ids), NBR_CONCURRENT_REQUESTS)

        with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
            index_by_future = {
                executor.submit(self._delete_one, asset_id): index for index, asset_id in enumerate(asset_ids)
            }
            for future in tqdm(
                concurrent.futures.as_completed(index_by_future), desc=f"Deleting from {self._endpoint}"
            ):
                index = index_by_future[future]
                asset_id = asset_ids[index]
                response = future.result()
                if not response.status_code == 200:
                    raise ValueError(
                        f"Error when deleting asset: {self._endpoint}/{asset_id}. {response.status_code=} {response.text=}"
                    )
                responses[index] = response
        return responses


class SequentialGetter:
    def __init__(self, session: requests.Session, endpoint: str):
        self._session = session
        self._endpoint = endpoint

    def __call__(self) -> List[Dict]:
        base_url, slug = self._get_base_url()

        resp = self._session.get(base_url + slug)
        if not resp.status_code == 200:
            raise RuntimeError(f"GET from {base_url+slug} failed with {resp.status_code}, {resp.text}.")

        resource_list = resp.json()

        while "next" in resp.links:
            slug = resp.links["next"]["url"]
            resp = self._session.get(base_url + slug)
            if not resp.status_code == 200:
                raise RuntimeError(f"GET from {base_url+slug} failed with {resp.status_code}, {resp.text}.")
            resource_list.extend(resp.json())

        return resource_list

    def _get_base_url(self) -> Tuple[str, str]:
        if ":5000" in self._endpoint:
            # Running locally
            parts = self._endpoint.split(":5000")
            base_url, slug = parts
            base_url += ":5000"
        else:
            parts = self._endpoint.split(".com")
            base_url, slug = parts
            base_url += ".com"
        return base_url, slug


def get_session_that_retries() -> requests.Session:
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    session.mount("http://", HTTPAdapter(max_retries=retries))
    return session



import time
from typing import Dict, List

from nyckel import OAuth2Renewer
from nyckel.functions.classification.classification import ClassificationFunctionURLHandler, ClassificationLabel
from nyckel.functions.utils import strip_nyckel_prefix
from nyckel.request_utils import ParallelDeleter, ParallelPoster, SequentialGetter, get_session_that_retries


class ClassificationLabelHandler:
    def __init__(self, function_id: str, auth: OAuth2Renewer):
        self._function_id = function_id
        self._auth = auth
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self._session = get_session_that_retries()

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})

    def create_labels(self, labels: List[ClassificationLabel]) -> List[str]:
        if len(labels) == 0:
            return []
        self._refresh_auth_token()
        bodies = [
            {"name": label.name, "description": label.description, "metadata": label.metadata} for label in labels
        ]
        responses = ParallelPoster(self._session, self._url_handler.api_endpoint(path="labels"))(bodies)

        label_ids = [strip_nyckel_prefix(resp.json()["id"]) for resp in responses]

        # Before returning, make sure the assets are available via the API.
        label_names_post_complete = False
        while not label_names_post_complete:
            print("Waiting to confirm label assets are available...")
            time.sleep(0.5)
            labels_retrieved = self.list_labels()
            label_names_post_complete = set([l.name for l in labels]).issubset([l.name for l in labels_retrieved])

        return label_ids

    def list_labels(self) -> List[ClassificationLabel]:
        self._refresh_auth_token()
        labels_dict_list = SequentialGetter(self._session, self._url_handler.api_endpoint(path="labels"))()
        labels = [self._label_from_dict(entry) for entry in labels_dict_list]
        return labels

    def read_label(self, label_id: str) -> ClassificationLabel:
        self._refresh_auth_token()
        response = self._session.get(self._url_handler.api_endpoint(path=f"labels/{label_id}"))
        if not response.status_code == 200:
            raise RuntimeError(
                f"Unable to fetch label {label_id} from {self._url_handler.train_page} {response.text=} {response.status_code=}"
            )
        label_dict = response.json()
        return self._label_from_dict(label_dict)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        raise NotImplementedError

    def delete_label(self, label_id: str) -> None:
        self._refresh_auth_token()
        response = self._session.delete(self._url_handler.api_endpoint(path=f"labels/{strip_nyckel_prefix(label_id)}"))
        assert response.status_code == 200, f"Delete failed with {response.status_code=}, {response.text=}"

    def delete_labels(self, label_ids: List[str]) -> None:
        if len(label_ids) == 0:
            return None
        self._refresh_auth_token()
        label_ids = [strip_nyckel_prefix(label_id) for label_id in label_ids]
        parallel_deleter = ParallelDeleter(self._session, self._url_handler.api_endpoint(path="labels"))
        parallel_deleter(label_ids)

    def _label_from_dict(self, label_dict: Dict) -> ClassificationLabel:
        return ClassificationLabel(
            name=label_dict["name"],
            id=strip_nyckel_prefix(label_dict["id"]),
            description=label_dict["description"] if "description" in label_dict else None,
            metadata=label_dict["metadata"] if "metadata" in label_dict else None,
        )
